library ArcNeuralNetwork.Neuron;

//Imports

import "dart:math";
import "layer.dart";
import 'Functions/inputfunction.dart';
import 'Functions/activationfunction.dart';

/*
Class Neuron: represent a neuron
      Parameters:
        - Id: identifier for the neuron --> String
        - InputConnection: a list of all the input connections --> List<Connection>
        - OutputConnections: a list of all output connections --> List<Connection>
        - Input: value of the input --> double
        - Output: value of the output --> double
        - Error: local error of the neuron --> double
        - Activation function --> ActivationFunction
        - Input Function: to calculate the input value --> InputFunction
      Methods:
        - calculateOutput: calculates the output of the neuron in base to the input and activation function.
        - calculateInput: calculates the input.
        - Check connections: boolean methods to know if there are connections set.
        - Add connection: add connections to neuron.
        - Remove connection: remove connection from neuron.
*/

class Neuron {

  String id;
  List<Connection> inputConnection;
  List<Connection> outputConnection;
  double input = 0.0;
  double output = 0.0;
  double error = 0.0;
  ActivationFunction activationFunction;
  InputFunction inputFunction;
  Neuron umbral;

  //////////////////////////////////////////////
  ////////////// Constructors //////////////////
  //////////////////////////////////////////////

  //Create a simple neuron with a stp activation function and a weight combination function for the input

  Neuron(this.id, {InputFunction inputFunction, ActivationFunction activationFunction, bool umbral}) {
    if(inputFunction != null){
      this.inputFunction = inputFunction;
    }else{
      this.inputFunction = new WeightCombination();
    }
    if(activationFunction != null){
      this.activationFunction = activationFunction;
    }else{
      this.activationFunction = new Step();
    }
    if(umbral!=null && umbral == true){
      this.createUmbral();
    }
    this.inputConnection = [];
    this.outputConnection = [];
  }

  Neuron.inputNeuron(this.id){
    this.inputConnection = [];
    this.outputConnection = [];
  }




  //Methods

  //calculates the output of the neuron in base to the input and activation function

  void createUmbral(){
    if(this.umbral!=null){
      Neuron umbral = new Neuron.inputNeuron("UmbralOf"+this.id);
      umbral.input = 1.0;
      umbral.inputConnection = null;
      Connection temp = new Connection(umbral,this);
      umbral.addOutputConnection(temp);
      this.addInputConnection(temp);
    }
  }

  void calculateOutput() {
    if(this.inputFunction == null && this.activationFunction == null && this.inputConnection.isEmpty){
      this.output = this.input;
    }else{
      this.output = this.activationFunction.getOutput(this.input);
    }

  }

  void calculateInput(){
    if (this.inputConnection.isNotEmpty) {
      this.input = this.inputFunction.getOutput(this.inputConnection);
    }
  }

  ///////////////////////////////////  Check connections  /////////////////////////////////

  bool get hasInputConnection {
    if (this.inputConnection.length > 0) {
      return true;
    } else {
      return false;
    }
  }

  bool get hasOutputConnection {
    if (this.outputConnection.length > 0) {
      return true;
    } else {
      return false;
    }
  }

  bool hasInputConnectionFrom(Neuron neuron) {
    return this.inputConnection.contains(neuron);
  }

  bool hasOutputConnectionTo(Neuron neuron) {
    return this.outputConnection.contains(neuron);
  }

  ///////////////////////////////////  Adding connections  /////////////////////////////////

  addInputConnection(Connection connection) {
    if (connection == null) {
      throw ("Attempt to add null connection to neuron!");
    }

    // make sure that connection instance is pointing to this neuron
    if (connection.to != this) {
      throw ("Cannot add input connection - bad fromNeuron specified!");
    }

    // if this neuron is already connected to neuron specified in connection do nothing
    if (this.hasInputConnectionFrom(connection.from)) {
      return;
    }
    this.inputConnection.add(connection);
    Neuron fromNeuron = connection.from;
    fromNeuron.addOutputConnection(connection);
  }

  addOutputConnection(Connection connection) {

    if (connection == null) {
      throw ("Attempt to add null connection to neuron!");
    }

    // make sure that connection instance is pointing to this neuron
    if (connection.from != this) {
      throw ("Cannot add output connection - bad fromNeuron specified!");
    }

    // if this neuron is already connected to neuron specified in connection do nothing
    if (this.hasOutputConnectionTo(connection.to)) {
      return;
    }
    this.outputConnection.add(connection);
  }

  ///////////////////////////////////  Remove connections  /////////////////////////////////

  removeInputConnection(Connection connection) {
    this.inputConnection.remove(connection);
  }

  removeOutputConnection(Connection connection) {
    this.outputConnection.remove(connection);
  }

  removeInputConnectionWithNeuronFrom(Neuron neuron) {
    Connection connection = this.inputConnectionNeuronFrom(neuron);
    this.removeInputConnection(connection);
  }

  removeOutputConnectionWithNeuronTo(Neuron neuron) {
    Connection connection = this.outputConnectionNeuronTo(neuron);
    this.removeOutputConnection(connection);
  }

  removeAllInputConnections() {
    this.inputConnection = [];
  }

  removeAllOutputConnections() {
    this.outputConnection = [];
  }

  removeAllConnections() {
    removeAllInputConnections();
    removeAllOutputConnections();
  }

  ///////////////////////////////////  Get input connections  ////////////////////////////////////

  Connection inputConnectionNeuronFrom(Neuron neuron) {
    for (var connection in this.inputConnection) {
      if (connection.from == neuron) return connection;
    }
    return null;
  }

  ///////////////////////////////////  Get output connections  ////////////////////////////////////

  Connection outputConnectionNeuronTo(Neuron neuron) {
    for (var connection in this.outputConnection) {
      if (connection.to == neuron) return connection;
    }
    return null;
  }

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////           CLASS CONNECTION                       //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
Class Connection: represent a connection between two neurons
      Parameters:
        - From: neuron where connection starts --> Neuron
        - To: neuron where connection ends -->Neuron
        - Weight: weight associated to the connection -->Weight
      Methods:
        - inputValue: calculates the input of the connection. This is important because that input value is used to calculate the total input of a neuron.
*/

class Connection {

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      PARAMETERS        ////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  Neuron from;
  Neuron to;
  Weight weight;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      CONSTRUCTORS      ////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


  Connection(Neuron from, Neuron to, [double value]) {
    if(from != null && to != null){
      if(value != null){
        this.weight = new Weight(value);
      }else{
        this.weight = new Weight();
      }
      this.from = from;
      this.to = to;
    }else{
      throw ("Neuron from and neuron to can't be null");
    }
  }

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      GETTERS AND SETTERS  /////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  /*
  * NAME: inputValue
  * RETURNS: the value of the output of the neuron from
  * PARAMETERS:nothing
  */

  double get inputValue => this.from.output;

  /*
  * NAME: weightValue
  * RETURNS: the value of the weight
  * PARAMETERS:nothing
  */

  double get weightValue => this.weight.value;

  /*
  * NAME: weightValue
  * RETURNS: nothing
  * PARAMETERS:
  * * double value -- new value to be set
  */

  void set weightValue(double value){this.weight.value = value;}

  /*
  * NAME: weightPreviousValue
  * RETURNS: the previous value of the weight
  * PARAMETERS:
  */

  double get weightPreviousValue => this.weight.previousValue;

  /*
  * NAME: weightVariation
  * RETURNS: the variation of the weight
  * PARAMETERS:nothing
  */

  double get weightVariation => this.weight.variation;

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////           CLASS WEIGHT                           //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

/*
Class Weight: represent the weight of a connection
      Parameters:
        - Value: the value of the weight at this moment -->double
        - PreviousValue: the value of the weight at the moment before -->double
        - Variation: the difference between the value and the previous value --> double
      Methods:
        - randomValue: generates a random number or a random number inside a range
        - increment: increments the value of the weight and calculates the variation.
        - decrement: decrements the value of the weight and calculates the variation.
*/

class Weight {

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      PARAMETERS        ////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


  double _value;
  double _previousValue;
  double _variation;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      CONSTRUCTORS      ////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  Weight([double value]) {
    if(value != null){
      this.value = value;
    }else{
      this.value = randomValue();
    }

    this._variation = 0.0;
  }

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      GETTERS AND SETTERS  /////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


  /*
  * NAME: value
  * RETURNS: nothing
  * PARAMETERS:
  * * double value -- new value to be set
  */

  void set value(double value) {
    if (this._value == null) {
      this._value = value;
    } else {
      this._previousValue = this._value;
      this._value = value;
      this._variation = (this._value - this._previousValue).abs();
    }
  }

  /*
  * NAME: value
  * RETURNS: the double value of the weight
  * PARAMETERS: nothing
  */

  double get value => this._value;

  /*
  * NAME: previousValue
  * RETURNS: the double previous value of the weight
  * PARAMETERS: nothing
  */

  double get previousValue => this._previousValue;

  /*
  * NAME: variation
  * RETURNS: the double value variation of the weight
  * PARAMETERS: nothing
  */

  double get variation => this._variation;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////      METHODS              /////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

  /*
  * NAME: decrement
  * RETURNS: double random value
  * PARAMETERS:nothing
  */

  double randomValue() {
    Random r = new Random();
    return r.nextDouble();
  }

  /*
  * NAME: increment
  * RETURNS: nothing
  * PARAMETERS:
  * * double value -- the increment value
  */

  void increment(double value) {
    this.value = this._value + value;
  }

  /*
  * NAME: decrement
  * RETURNS: nothing
  * PARAMETERS:
  * * double value -- the decrement value
  */

  void decrement(double value) {
    this.value = this._value - value;
  }

}
